import httpx

# default coze base_url is api.coze.com
COZE_COM_BASE_URL = "https://api.coze.com"
# support change to api.coze.cn
COZE_CN_BASE_URL = "https://api.coze.cn"

# default timeout is 10 minutes, with 5 seconds connect timeout
DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
DEFAULT_CONNECTION_LIMITS = httpx.Limits(max_connections=1000, max_keepalive_connections=100)



from .audio.rooms import CreateRoomResult
from .audio.speech import AudioFormat
from .audio.voices import Voice
from .auth import (
    AsyncDeviceOAuthApp,
    AsyncJWTOAuthApp,
    AsyncPKCEOAuthApp,
    AsyncWebOAuthApp,
    Auth,
    DeviceAuthCode,
    DeviceOAuthApp,
    JWTAuth,
    JWTOAuthApp,
    OAuthApp,
    OAuthToken,
    PKCEOAuthApp,
    Scope,
    TokenAuth,
    WebOAuthApp,
)
from .bots import (
    Bot,
    BotModelInfo,
    BotOnboardingInfo,
    BotPluginAPIInfo,
    BotPluginInfo,
    BotPromptInfo,
    SimpleBot,
)
from .chat import (
    Chat,
    ChatError,
    ChatEvent,
    ChatEventType,
    ChatPoll,
    ChatRequiredAction,
    ChatRequiredActionType,
    ChatStatus,
    ChatSubmitToolOutputs,
    ChatToolCall,
    ChatToolCallFunction,
    ChatToolCallType,
    ChatUsage,
    Message,
    MessageContentType,
    MessageObjectString,
    MessageObjectStringType,
    MessageRole,
    MessageType,
    ToolOutput,
)
from .config import (
    COZE_CN_BASE_URL,
    COZE_COM_BASE_URL,
    DEFAULT_CONNECTION_LIMITS,
    DEFAULT_TIMEOUT,
)
from .conversations import Conversation, Section
from .coze import AsyncCoze, Coze
from .exception import CozeAPIError, CozeError, CozeInvalidEventError, CozePKCEAuthError, CozePKCEAuthErrorType
from .files import File
from .knowledge.documents import (
    Document,
    DocumentBase,
    DocumentChunkStrategy,
    DocumentFormatType,
    DocumentSourceInfo,
    DocumentSourceType,
    DocumentStatus,
    DocumentUpdateRule,
    DocumentUpdateType,
)
from .log import setup_logging
from .model import (
    AsyncLastIDPaged,
    AsyncNumberPaged,
    AsyncPagedBase,
    AsyncStream,
    FileHTTPResponse,
    LastIDPaged,
    LastIDPagedResponse,
    NumberPaged,
    NumberPagedResponse,
    Stream,
)
from .request import AsyncHTTPClient, SyncHTTPClient
from .version import VERSION
from .workflows.runs import (
    WorkflowEvent,
    WorkflowEventError,
    WorkflowEventInterrupt,
    WorkflowEventInterruptData,
    WorkflowEventMessage,
    WorkflowEventType,
    WorkflowRunResult,
)
from .workflows.runs.run_histories import WorkflowExecuteStatus, WorkflowRunHistory, WorkflowRunMode
from .workspaces import Workspace, WorkspaceRoleType, WorkspaceType

__all__ = [
    "VERSION",
    # audio.rooms
    "CreateRoomResult",
    # audio.voices
    "Voice",
    "AudioFormat",
    # auth
    "AsyncDeviceOAuthApp",
    "AsyncJWTOAuthApp",
    "AsyncPKCEOAuthApp",
    "AsyncWebOAuthApp",
    "Auth",
    "DeviceAuthCode",
    "DeviceOAuthApp",
    "JWTAuth",
    "JWTOAuthApp",
    "OAuthApp",
    "OAuthToken",
    "PKCEOAuthApp",
    "Scope",
    "TokenAuth",
    "WebOAuthApp",
    # bots
    "BotPromptInfo",
    "BotOnboardingInfo",
    "BotModelInfo",
    "BotPluginAPIInfo",
    "BotPluginInfo",
    "Bot",
    "SimpleBot",
    # chat
    "MessageRole",
    "MessageType",
    "MessageContentType",
    "MessageObjectStringType",
    "MessageObjectString",
    "Message",
    "ChatStatus",
    "ChatError",
    "ChatRequiredActionType",
    "ChatToolCallType",
    "ChatToolCallFunction",
    "ChatToolCall",
    "ChatSubmitToolOutputs",
    "ChatRequiredAction",
    "ChatUsage",
    "Chat",
    "ChatPoll",
    "ChatEventType",
    "ChatEvent",
    "ToolOutput",
    # conversations
    "Conversation",
    "Section",
    # files
    "File",
    # knowledge.documents
    "DocumentChunkStrategy",
    "DocumentFormatType",
    "DocumentSourceType",
    "DocumentStatus",
    "DocumentUpdateType",
    "Document",
    "DocumentSourceInfo",
    "DocumentUpdateRule",
    "DocumentBase",
    # workflows.runs
    "WorkflowRunResult",
    "WorkflowEventType",
    "WorkflowEventMessage",
    "WorkflowEventInterruptData",
    "WorkflowEventInterrupt",
    "WorkflowEventError",
    "WorkflowEvent",
    # workflows.runs.run_histories
    "WorkflowExecuteStatus",
    "WorkflowRunMode",
    "WorkflowRunHistory",
    # workspaces
    "WorkspaceRoleType",
    "WorkspaceType",
    "Workspace",
    # log
    "setup_logging",
    # config
    "COZE_COM_BASE_URL",
    "COZE_CN_BASE_URL",
    "DEFAULT_TIMEOUT",
    "DEFAULT_CONNECTION_LIMITS",
    # coze
    "AsyncCoze",
    "Coze",
    # exception
    "CozeError",
    "CozeAPIError",
    "CozeInvalidEventError",
    "CozePKCEAuthError",
    "CozePKCEAuthErrorType",
    # model
    "AsyncLastIDPaged",
    "AsyncNumberPaged",
    "AsyncPagedBase",
    "AsyncStream",
    "FileHTTPResponse",
    "LastIDPaged",
    "LastIDPagedResponse",
    "NumberPaged",
    "NumberPagedResponse",
    "Stream",
    # request
    "SyncHTTPClient",
    "AsyncHTTPClient",
]



from typing import Dict, List, Optional

from cozepy.auth import Auth
from cozepy.chat import Message, MessageContentType, MessageRole
from cozepy.model import AsyncLastIDPaged, CozeModel, HTTPRequest, LastIDPaged, LastIDPagedResponse
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash


class _PrivateListMessageResp(CozeModel, LastIDPagedResponse[Message]):
    first_id: str
    last_id: str
    has_more: bool
    items: List[Message]

    def get_first_id(self) -> str:
        return self.first_id

    def get_last_id(self) -> str:
        return self.last_id

    def get_has_more(self) -> bool:
        return self.has_more

    def get_items(self) -> List[Message]:
        return self.items


class MessagesClient(object):
    """
    Message class.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    def create(
        self,
        *,
        conversation_id: str,
        role: MessageRole,
        content: str,
        content_type: MessageContentType,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Message:
        """
        Create a message and add it to the specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/create_message
        docs cn: https://www.coze.cn/docs/developer_guides/create_message

        :param conversation_id: The ID of the conversation.
        :param role: The entity that sent this message.
        :param content: The content of the message, supporting pure text, multimodal (mixed input of text, images,
        files), cards, and various types of content.
        :param content_type: The type of message content.
        :param meta_data: Additional information when creating a message, and this additional information will also be
        returned when retrieving messages.
        :return:
        """
        url = f"{self._base_url}/v1/conversation/message/create"
        params = {
            "conversation_id": conversation_id,
        }
        body = {
            "role": role,
            "content": content,
            "content_type": content_type,
            "meta_data": meta_data,
        }

        return self._requester.request("post", url, False, Message, params=params, body=body)

    def list(
        self,
        *,
        conversation_id: str,
        order: str = "desc",
        chat_id: Optional[str] = None,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        limit: int = 50,
    ) -> LastIDPaged[Message]:
        """
        Get the message list of a specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/list_message
        docs zh: https://www.coze.cn/docs/developer_guides/list_message

        :param conversation_id: The ID of the conversation.
        :param order: The sorting method for the message list.
        :param chat_id: The ID of the Chat.
        :param before_id: Get messages before the specified position.
        :param after_id: Get messages after the specified position.
        :param limit: The amount of data returned per query. Default is 50, with a range of 1 to 50.
        :return: The message list of the specified conversation.
        """
        url = f"{self._base_url}/v1/conversation/message/list"
        params = {
            "conversation_id": conversation_id,
        }

        def request_maker(i_before_id: str, i_after_id: str) -> HTTPRequest:
            return self._requester.make_request(
                "POST",
                url,
                json={
                    "order": order,
                    "chat_id": chat_id,
                    "before_id": i_before_id if i_before_id else None,
                    "after_id": i_after_id if i_after_id else None,
                    "limit": limit,
                },
                params=params,
                cast=_PrivateListMessageResp,
                is_async=False,
                stream=False,
            )

        return LastIDPaged(
            before_id=before_id or "",
            after_id=after_id or "",
            requestor=self._requester,
            request_maker=request_maker,
        )

    def retrieve(
        self,
        *,
        conversation_id: str,
        message_id: str,
    ) -> Message:
        """
        Get the detailed information of specified message.

        docs en: https://www.coze.com/docs/developer_guides/retrieve_message
        docs zh: https://www.coze.cn/docs/developer_guides/retrieve_message

        :param conversation_id: The ID of the conversation.
        :param message_id: The ID of the message.
        :return: The detailed information of specified message.
        """
        url = f"{self._base_url}/v1/conversation/message/retrieve"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }

        return self._requester.request("get", url, False, Message, params=params)

    def update(
        self,
        *,
        conversation_id: str,
        message_id: str,
        content: Optional[str] = None,
        content_type: Optional[MessageContentType] = None,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Message:
        """
        Modify a message, supporting the modification of message content, additional content, and message type.

        docs en: https://www.coze.com/docs/developer_guides/modify_message
        docs cn: https://www.coze.cn/docs/developer_guides/modify_message

        :param conversation_id: The ID of the conversation.
        :param message_id: The ID of the message.
        :param content: The content of the message, supporting pure text, multimodal (mixed input of text, images,
        files), cards, and various types of content.
        :param content_type: The type of message content.
        :param meta_data:
        :return: The detailed information of specified message.
        """
        url = f"{self._base_url}/v1/conversation/message/modify"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }
        body = {
            "content": content,
            "content_type": content_type,
            "meta_data": meta_data,
        }

        return self._requester.request("post", url, False, Message, params=params, body=body, data_field="message")

    def delete(
        self,
        *,
        conversation_id: str,
        message_id: str,
    ) -> Message:
        """
        Call the API to delete a message within a specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/delete_message
        docs zh: https://www.coze.cn/docs/developer_guides/delete_message

        :param conversation_id:
        :param message_id:
        :return:
        """
        url = f"{self._base_url}/v1/conversation/message/delete"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }

        return self._requester.request("post", url, False, Message, params=params)


class AsyncMessagesClient(object):
    """
    Message class.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    async def create(
        self,
        *,
        conversation_id: str,
        role: MessageRole,
        content: str,
        content_type: MessageContentType,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Message:
        """
        Create a message and add it to the specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/create_message
        docs cn: https://www.coze.cn/docs/developer_guides/create_message

        :param conversation_id: The ID of the conversation.
        :param role: The entity that sent this message.
        :param content: The content of the message, supporting pure text, multimodal (mixed input of text, images,
        files), cards, and various types of content.
        :param content_type: The type of message content.
        :param meta_data: Additional information when creating a message, and this additional information will also be
        returned when retrieving messages.
        :return:
        """
        url = f"{self._base_url}/v1/conversation/message/create"
        params = {
            "conversation_id": conversation_id,
        }
        body = {
            "role": role,
            "content": content,
            "content_type": content_type,
            "meta_data": meta_data,
        }

        return await self._requester.arequest("post", url, False, Message, params=params, body=body)

    async def list(
        self,
        *,
        conversation_id: str,
        order: str = "desc",
        chat_id: Optional[str] = None,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        limit: int = 50,
    ) -> AsyncLastIDPaged[Message]:
        """
        Get the message list of a specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/list_message
        docs zh: https://www.coze.cn/docs/developer_guides/list_message

        :param conversation_id: The ID of the conversation.
        :param order: The sorting method for the message list.
        :param chat_id: The ID of the Chat.
        :param before_id: Get messages before the specified position.
        :param after_id: Get messages after the specified position.
        :param limit: The amount of data returned per query. Default is 50, with a range of 1 to 50.
        :return: The message list of the specified conversation.
        """
        url = f"{self._base_url}/v1/conversation/message/list"
        params = {
            "conversation_id": conversation_id,
        }

        def request_maker(i_before_id: str, i_after_id: str) -> HTTPRequest:
            return self._requester.make_request(
                "POST",
                url,
                json={
                    "order": order,
                    "chat_id": chat_id,
                    "before_id": i_before_id if i_before_id else None,
                    "after_id": i_after_id if i_after_id else None,
                    "limit": limit,
                },
                params=params,
                cast=_PrivateListMessageResp,
                is_async=False,
                stream=False,
            )

        return await AsyncLastIDPaged.build(
            before_id=before_id or "",
            after_id=after_id or "",
            requestor=self._requester,
            request_maker=request_maker,
        )

    async def retrieve(
        self,
        *,
        conversation_id: str,
        message_id: str,
    ) -> Message:
        """
        Get the detailed information of specified message.

        docs en: https://www.coze.com/docs/developer_guides/retrieve_message
        docs zh: https://www.coze.cn/docs/developer_guides/retrieve_message

        :param conversation_id: The ID of the conversation.
        :param message_id: The ID of the message.
        :return: The detailed information of specified message.
        """
        url = f"{self._base_url}/v1/conversation/message/retrieve"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }

        return await self._requester.arequest("get", url, False, Message, params=params)

    async def update(
        self,
        *,
        conversation_id: str,
        message_id: str,
        content: Optional[str] = None,
        content_type: Optional[MessageContentType] = None,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Message:
        """
        Modify a message, supporting the modification of message content, additional content, and message type.

        docs en: https://www.coze.com/docs/developer_guides/modify_message
        docs cn: https://www.coze.cn/docs/developer_guides/modify_message

        :param conversation_id: The ID of the conversation.
        :param message_id: The ID of the message.
        :param content: The content of the message, supporting pure text, multimodal (mixed input of text, images,
        files), cards, and various types of content.
        :param content_type: The type of message content.
        :param meta_data:
        :return: The detailed information of specified message.
        """
        url = f"{self._base_url}/v1/conversation/message/modify"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }
        body = {
            "content": content,
            "content_type": content_type,
            "meta_data": meta_data,
        }

        return await self._requester.arequest(
            "post", url, False, Message, params=params, body=body, data_field="message"
        )

    async def delete(
        self,
        *,
        conversation_id: str,
        message_id: str,
    ) -> Message:
        """
        Call the API to delete a message within a specified conversation.

        docs en: https://www.coze.com/docs/developer_guides/delete_message
        docs zh: https://www.coze.cn/docs/developer_guides/delete_message

        :param conversation_id:
        :param message_id:
        :return:
        """
        url = f"{self._base_url}/v1/conversation/message/delete"
        params = {
            "conversation_id": conversation_id,
            "message_id": message_id,
        }

        return await self._requester.arequest("post", url, False, Message, params=params)



import base64
import hashlib
import random
import sys
import wave

if sys.version_info < (3, 10):

    async def anext(iterator, default=None):
        try:
            return await iterator.__anext__()
        except StopAsyncIteration:
            if default is not None:
                return default
            raise
else:
    from builtins import anext

    _ = anext


def base64_encode_string(s: str) -> str:
    return base64.standard_b64encode(s.encode("utf-8")).decode("utf-8")


def random_hex(length):
    hex_characters = "0123456789abcdef"
    return "".join(random.choice(hex_characters) for _ in range(length))


def gen_s256_code_challenge(code_verifier):
    # 1. SHA256(ASCII(code_verifier))
    sha256_hash = hashlib.sha256(code_verifier.encode("ascii")).digest()
    # 2. BASE64URL-ENCODE
    code_challenge = base64.urlsafe_b64encode(sha256_hash).decode("ascii")
    # 3. remove =
    code_challenge = code_challenge.rstrip("=")
    return code_challenge


def remove_url_trailing_slash(base_url: str) -> str:
    if base_url:
        return base_url.rstrip("/")
    return base_url


def write_pcm_to_wav_file(
    pcm_data: bytes, filepath: str, channels: int = 1, sample_width: int = 2, frame_rate: int = 24000
):
    """
    Save PCM binary data to WAV file

    :param pcm_data: PCM binary data (24kHz, 16-bit, 1 channel, little-endian)
    :param output_filename: Output WAV filename
    """

    with wave.open(filepath, "wb") as wav_file:
        # Set WAV file parameters
        wav_file.setnchannels(channels)
        wav_file.setsampwidth(sample_width)
        wav_file.setframerate(frame_rate)

        # Write PCM data
        wav_file.writeframes(pcm_data)



from enum import IntEnum
from typing import List, Optional

from cozepy.auth import Auth
from cozepy.model import AsyncNumberPaged, CozeModel, NumberPaged, NumberPagedResponse
from cozepy.request import HTTPRequest, Requester
from cozepy.util import remove_url_trailing_slash


class BotPromptInfo(CozeModel):
    # Configured prompt of the bot.
    prompt: str


class BotOnboardingInfo(CozeModel):
    # Configured prologue of the bot.
    prologue: str = ""
    # The list of recommended questions configured for the bot.
    # This field will not be returned when user suggested questions are not enabled.
    suggested_questions: List[str] = []


class BotModelInfo(CozeModel):
    # The ID of the model.
    model_id: str
    # The name of the model.
    model_name: str


class BotMode(IntEnum):
    SINGL_EAGENT = 0
    MULTI_AGENT = 1
    SINGLE_AGENT_WORKFLOW = 2


class BotPluginAPIInfo(CozeModel):
    # The ID of the tool.
    api_id: str
    # The name of the tool.
    name: str
    # The description of the tool.
    description: str


class BotPluginInfo(CozeModel):
    # The unique identifier for the plugin.
    plugin_id: str
    # The name of the plugin.
    name: str
    # The description of the plugin.
    description: str
    # The avatar URL for the plugin.
    icon_url: str
    # The list of tool information for the plugin. For more information, see PluginAPI object.
    api_info_list: List[BotPluginAPIInfo]


class Bot(CozeModel):
    # The ID for the bot.
    bot_id: str
    # The name of the bot.
    name: Optional[str] = None
    # The description of the bot.
    description: Optional[str] = None
    # The URL address for the bot's avatar.
    icon_url: Optional[str] = None
    # The creation time, in the format of a 10-digit Unix timestamp in seconds (s).
    create_time: Optional[int] = None
    # The update time, in the format of a 10-digit Unix timestamp in seconds (s).
    update_time: Optional[int] = None
    # The latest version of the bot.
    version: Optional[str] = None
    # The prompt configuration for the bot. For more information, see Prompt object.
    prompt_info: Optional[BotPromptInfo] = None
    # The onboarding message configuration for the bot. For more information, see Onboarding object.
    onboarding_info: Optional[BotOnboardingInfo] = None
    # The mode of the Bot, values: 0: Single Agent mode, 1: Multi Agent mode, 3: Single Agent Workflow mode.
    bot_mode: Optional[BotMode] = None
    # The plugins configured for the bot. For more information, see  Plugin object.
    plugin_info_list: Optional[List[BotPluginInfo]] = None
    # The model configured for the bot. For more information, see Model object.
    model_info: Optional[BotModelInfo] = None


class SimpleBot(CozeModel):
    # The ID for the bot.
    bot_id: str
    # The name of the bot.
    bot_name: str
    # The description of the bot.
    description: str
    # The URL address for the bot's avatar.
    icon_url: str
    # The latest publish time of the bot, in the format of a 10-digit Unix timestamp in seconds (s).
    # This API returns the list of bots sorted in descending order by this field.
    publish_time: str


class _PrivateListBotsData(CozeModel, NumberPagedResponse[SimpleBot]):
    space_bots: List[SimpleBot]
    total: int

    def get_total(self) -> Optional[int]:
        return self.total

    def get_has_more(self) -> Optional[bool]:
        return None

    def get_items(self) -> List[SimpleBot]:
        return self.space_bots


class BotsClient(object):
    """
    Bot class.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    def create(
        self,
        *,
        space_id: str,
        name: str,
        description: Optional[str] = None,
        icon_file_id: Optional[str] = None,
        prompt_info: Optional[BotPromptInfo] = None,
        onboarding_info: Optional[BotOnboardingInfo] = None,
    ) -> Bot:
        url = f"{self._base_url}/v1/bot/create"
        body = {
            "space_id": space_id,
            "name": name,
            "description": description,
            "icon_file_id": icon_file_id,
            "prompt_info": prompt_info.model_dump() if prompt_info else None,
            "onboarding_info": onboarding_info.model_dump() if onboarding_info else None,
        }

        return self._requester.request("post", url, False, Bot, body=body)

    def update(
        self,
        *,
        bot_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        icon_file_id: Optional[str] = None,
        prompt_info: Optional[BotPromptInfo] = None,
        onboarding_info: Optional[BotOnboardingInfo] = None,
    ) -> None:
        """
        Update the configuration of a bot.
        This API can be used to update all bots created through the Coze platform or via the API.
        In addition to updating the bot's name and description, avatar, personality and reply logic,
        and opening remarks, this API also supports binding a knowledge base to the bot.

        docs en: https://www.coze.com/docs/developer_guides/update_bot
        docs zh: https://www.coze.cn/docs/developer_guides/update_bot

        :param bot_id: The ID of the bot that the API interacts with.
        :param name: The name of the bot. It should be 1 to 20 characters long.
        :param description: The description of the Bot. It can be 0 to 500 characters long. The default is empty.
        :param icon_file_id: The file ID for the Bot's avatar. If no file ID is specified, the Coze platform will
        assign a default avatar for the bot. To use a custom avatar, first upload the local file through the Upload
        file interface and obtain the file ID from the interface response.
        :param prompt_info: The personality and reply logic of the bot.
        :param onboarding_info: The settings related to the bot's opening remarks.
        :return: None
        """
        url = f"{self._base_url}/v1/bot/update"
        body = {
            "bot_id": bot_id,
            "name": name,
            "description": description,
            "icon_file_id": icon_file_id,
            "prompt_info": prompt_info.model_dump() if prompt_info else None,
            "onboarding_info": onboarding_info.model_dump() if onboarding_info else None,
        }

        return self._requester.request(
            "post",
            url,
            False,
            None,
            body=body,
        )

    def publish(self, *, bot_id: str, connector_ids: Optional[List[str]] = None, **kwargs) -> Bot:
        url = f"{self._base_url}/v1/bot/publish"
        headers: Optional[dict] = kwargs.get("headers")
        if not connector_ids:
            connector_ids = ["1024"]
        body = {
            "bot_id": bot_id,
            "connector_ids": connector_ids,
        }

        return self._requester.request("post", url, False, Bot, headers=headers, body=body)

    def retrieve(self, *, bot_id: str) -> Bot:
        """
        Get the configuration information of the bot, which must have been published
        to the Bot as API channel.
        获取指定 Bot 的配置信息，此 Bot 必须已发布到 Bot as API 渠道中。

        docs en: https://www.coze.com/docs/developer_guides/get_metadata
        docs zh: https://www.coze.cn/docs/developer_guides/get_metadata

        :param bot_id: The ID of the bot that the API interacts with.
        要查看的 Bot ID。
        :return: bot object
        Bot 的配置信息
        """
        url = f"{self._base_url}/v1/bot/get_online_info"
        params = {"bot_id": bot_id}

        return self._requester.request("get", url, False, Bot, params=params)

    def list(self, *, space_id: str, page_num: int = 1, page_size: int = 20) -> NumberPaged[SimpleBot]:
        """
        Get the bots published as API service.
        查看指定空间发布到 Bot as API 渠道的 Bot 列表。

        docs en: https://www.coze.com/docs/developer_guides/published_bots_list
        docs zh: https://www.coze.cn/docs/developer_guides/published_bots_list

        :param space_id: The ID of the space.
        Bot 所在的空间的 Space ID。Space ID 是空间的唯一标识。
        :param page_num: Pagination size. The default is 20, meaning that 20 data entries are returned per page.
        分页大小。默认为 20，即每页返回 20 条数据。
        :param page_size: Page number for paginated queries. The default is 1,
        meaning that the data return starts from the first page.
        分页查询时的页码。默认为 1，即从第一页数据开始返回。
        :return: Specify the list of Bots published to the Bot as an API channel in space.
        指定空间发布到 Bot as API 渠道的 Bot 列表。
        """
        url = f"{self._base_url}/v1/space/published_bots_list"

        def request_maker(i_page_num: int, i_page_size: int) -> HTTPRequest:
            return self._requester.make_request(
                "GET",
                url,
                params={
                    "space_id": space_id,
                    "page_size": i_page_size,
                    "page_index": i_page_num,
                },
                cast=_PrivateListBotsData,
                is_async=False,
                stream=False,
            )

        return NumberPaged(
            page_num=page_num,
            page_size=page_size,
            requestor=self._requester,
            request_maker=request_maker,
        )


class AsyncBotsClient(object):
    """
    Bot class.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    async def create(
        self,
        *,
        space_id: str,
        name: str,
        description: Optional[str] = None,
        icon_file_id: Optional[str] = None,
        prompt_info: Optional[BotPromptInfo] = None,
        onboarding_info: Optional[BotOnboardingInfo] = None,
    ) -> Bot:
        url = f"{self._base_url}/v1/bot/create"
        body = {
            "space_id": space_id,
            "name": name,
            "description": description,
            "icon_file_id": icon_file_id,
            "prompt_info": prompt_info.model_dump() if prompt_info else None,
            "onboarding_info": onboarding_info.model_dump() if onboarding_info else None,
        }

        return await self._requester.arequest("post", url, False, Bot, body=body)

    async def update(
        self,
        *,
        bot_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        icon_file_id: Optional[str] = None,
        prompt_info: Optional[BotPromptInfo] = None,
        onboarding_info: Optional[BotOnboardingInfo] = None,
    ) -> None:
        """
        Update the configuration of a bot.
        This API can be used to update all bots created through the Coze platform or via the API.
        In addition to updating the bot's name and description, avatar, personality and reply logic,
        and opening remarks, this API also supports binding a knowledge base to the bot.

        docs en: https://www.coze.com/docs/developer_guides/update_bot
        docs zh: https://www.coze.cn/docs/developer_guides/update_bot

        :param bot_id: The ID of the bot that the API interacts with.
        :param name: The name of the bot. It should be 1 to 20 characters long.
        :param description: The description of the Bot. It can be 0 to 500 characters long. The default is empty.
        :param icon_file_id: The file ID for the Bot's avatar. If no file ID is specified, the Coze platform will
        assign a default avatar for the bot. To use a custom avatar, first upload the local file through the Upload
        file interface and obtain the file ID from the interface response.
        :param prompt_info: The personality and reply logic of the bot.
        :param onboarding_info: The settings related to the bot's opening remarks.
        :return: None
        """
        url = f"{self._base_url}/v1/bot/update"
        body = {
            "bot_id": bot_id,
            "name": name,
            "description": description,
            "icon_file_id": icon_file_id,
            "prompt_info": prompt_info.model_dump() if prompt_info else None,
            "onboarding_info": onboarding_info.model_dump() if onboarding_info else None,
        }

        return await self._requester.arequest("post", url, False, None, body=body)

    async def publish(
        self,
        *,
        bot_id: str,
        connector_ids: Optional[List[str]] = None,
        **kwargs,
    ) -> Bot:
        url = f"{self._base_url}/v1/bot/publish"
        headers: Optional[dict] = kwargs.get("headers")
        if not connector_ids:
            connector_ids = ["1024"]
        body = {
            "bot_id": bot_id,
            "connector_ids": connector_ids,
        }

        return await self._requester.arequest("post", url, False, Bot, headers=headers, body=body)

    async def retrieve(self, *, bot_id: str) -> Bot:
        """
        Get the configuration information of the bot, which must have been published
        to the Bot as API channel.
        获取指定 Bot 的配置信息，此 Bot 必须已发布到 Bot as API 渠道中。

        docs en: https://www.coze.com/docs/developer_guides/get_metadata
        docs zh: https://www.coze.cn/docs/developer_guides/get_metadata

        :param bot_id: The ID of the bot that the API interacts with.
        要查看的 Bot ID。
        :return: bot object
        Bot 的配置信息
        """
        url = f"{self._base_url}/v1/bot/get_online_info"
        params = {"bot_id": bot_id}

        return await self._requester.arequest("get", url, False, Bot, params=params)

    async def list(self, *, space_id: str, page_num: int = 1, page_size: int = 20) -> AsyncNumberPaged[SimpleBot]:
        """
        Get the bots published as API service.
        查看指定空间发布到 Bot as API 渠道的 Bot 列表。

        docs en: https://www.coze.com/docs/developer_guides/published_bots_list
        docs zh: https://www.coze.cn/docs/developer_guides/published_bots_list

        :param space_id: The ID of the space.
        Bot 所在的空间的 Space ID。Space ID 是空间的唯一标识。
        :param page_num: Pagination size. The default is 20, meaning that 20 data entries are returned per page.
        分页大小。默认为 20，即每页返回 20 条数据。
        :param page_size: Page number for paginated queries. The default is 1,
        meaning that the data return starts from the first page.
        分页查询时的页码。默认为 1，即从第一页数据开始返回。
        :return: Specify the list of Bots published to the Bot as an API channel in space.
        指定空间发布到 Bot as API 渠道的 Bot 列表。
        """
        url = f"{self._base_url}/v1/space/published_bots_list"

        def request_maker(i_page_num: int, i_page_size: int) -> HTTPRequest:
            return self._requester.make_request(
                "GET",
                url,
                params={
                    "space_id": space_id,
                    "page_size": i_page_size,
                    "page_index": i_page_num,
                },
                cast=_PrivateListBotsData,
                is_async=False,
                stream=False,
            )

        return await AsyncNumberPaged.build(
            page_num=page_num,
            page_size=page_size,
            requestor=self._requester,
            request_maker=request_maker,
        )



from enum import Enum
from typing import Optional


class CozeError(Exception):
    """
    base class for all coze errors
    """

    pass


class CozeAPIError(CozeError):
    def __init__(self, code: Optional[int] = None, msg: str = "", logid: Optional[str] = None):
        self.code = code
        self.msg = msg
        self.logid = logid
        if code and code > 0:
            super().__init__(f"code: {code}, msg: {msg}, logid: {logid}")
        else:
            super().__init__(f"msg: {msg}, logid: {logid}")


class CozePKCEAuthErrorType(str, Enum):
    AUTHORIZATION_PENDING = "authorization_pending"
    SLOW_DOWN = "slow_down"
    ACCESS_DENIED = "access_denied"
    EXPIRED_TOKEN = "expired_token"


COZE_PKCE_AUTH_ERROR_TYPE_ENUMS = set(e.value for e in CozePKCEAuthErrorType)


class CozePKCEAuthError(CozeError):
    def __init__(self, error: CozePKCEAuthErrorType, logid: Optional[str] = None):
        super().__init__(f"pkce auth error: {error.value}")
        self.error = error
        self.logid = logid


class CozeInvalidEventError(CozeError):
    def __init__(self, field: str = "", data: str = "", logid: str = ""):
        self.field = field
        self.data = data
        self.logid = logid
        if field:
            super().__init__(f"invalid event, field: {field}, data: {data}, logid: {logid}")
        else:
            super().__init__(f"invalid event, data: {data}, logid: {logid}")



from typing import TYPE_CHECKING, Optional

from cozepy.auth import Auth
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash

if TYPE_CHECKING:
    from .runs import AsyncWorkflowsRunsClient, WorkflowsRunsClient


class WorkflowsClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._runs: Optional[WorkflowsRunsClient] = None

    @property
    def runs(self) -> "WorkflowsRunsClient":
        if not self._runs:
            from .runs import WorkflowsRunsClient

            self._runs = WorkflowsRunsClient(self._base_url, self._auth, self._requester)
        return self._runs


class AsyncWorkflowsClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._runs: Optional[AsyncWorkflowsRunsClient] = None

    @property
    def runs(self) -> "AsyncWorkflowsRunsClient":
        if not self._runs:
            from .runs import AsyncWorkflowsRunsClient

            self._runs = AsyncWorkflowsRunsClient(self._base_url, self._auth, self._requester)
        return self._runs



from typing import TYPE_CHECKING, Optional

from cozepy.auth import Auth
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash

if TYPE_CHECKING:
    from .documents import AsyncDocumentsClient, DocumentsClient


class KnowledgeClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._documents: Optional[DocumentsClient] = None

    @property
    def documents(self) -> "DocumentsClient":
        if self._documents is None:
            from .documents import DocumentsClient

            self._documents = DocumentsClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._documents


class AsyncKnowledgeClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._documents: Optional[AsyncDocumentsClient] = None

    @property
    def documents(self) -> "AsyncDocumentsClient":
        if self._documents is None:
            from .documents import AsyncDocumentsClient

            self._documents = AsyncDocumentsClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._documents



import abc
import time
from typing import List, Optional
from urllib.parse import quote_plus, urlparse

from authlib.jose import jwt  # type: ignore
from typing_extensions import Literal

from cozepy.config import COZE_CN_BASE_URL, COZE_COM_BASE_URL
from cozepy.exception import CozePKCEAuthError, CozePKCEAuthErrorType
from cozepy.model import CozeModel
from cozepy.request import Requester
from cozepy.util import gen_s256_code_challenge, random_hex, remove_url_trailing_slash


class OAuthToken(CozeModel):
    # The requested access token. The app can use this token to authenticate to the Coze resource.
    access_token: str
    # How long the access token is valid, in seconds (UNIX timestamp)
    expires_in: int
    # An OAuth 2.0 refresh token. The app can use this token to acquire other access tokens after the current access token expires. Refresh tokens are long-lived.
    refresh_token: str = ""
    # fixed: Bearer
    token_type: str = ""


class DeviceAuthCode(CozeModel):
    # device code
    device_code: str
    # The user code
    user_code: str
    # The verification uri
    verification_uri: str
    # The interval of the polling request
    interval: int = 5
    # The expiration time of the device code
    expires_in: int
    verification_url: Optional[str] = None


class ScopeAccountPermission(CozeModel):
    permission_list: List[str]


class ScopeAttributeConstraintConnectorBotChatAttribute(CozeModel):
    bot_id_list: List[str]


class ScopeAttributeConstraint(CozeModel):
    connector_bot_chat_attribute: ScopeAttributeConstraintConnectorBotChatAttribute


class Scope(CozeModel):
    account_permission: Optional[ScopeAccountPermission] = None
    attribute_constraint: Optional[ScopeAttributeConstraint] = None

    @staticmethod
    def build_bot_chat(bot_id_list: List[str], permission_list: Optional[List[str]] = None) -> "Scope":
        if not permission_list:
            permission_list = ["Connector.botChat"]
        return Scope(
            account_permission=ScopeAccountPermission(permission_list=permission_list),
            attribute_constraint=ScopeAttributeConstraint(
                connector_bot_chat_attribute=ScopeAttributeConstraintConnectorBotChatAttribute(bot_id_list=bot_id_list)
            )
            if bot_id_list
            else None,
        )


class OAuthApp(object):
    def __init__(self, client_id: str, base_url: str, www_base_url: str):
        self._client_id = client_id
        self._base_url = remove_url_trailing_slash(base_url)
        self._api_endpoint = urlparse(base_url).netloc
        self._www_base_url = www_base_url
        self._requester = Requester()

    def _get_oauth_url(
        self,
        redirect_uri: str,
        code_challenge: Optional[str] = None,
        code_challenge_method: Optional[str] = None,
        state: str = "",
        workspace_id: Optional[str] = None,
    ):
        params = {
            "response_type": "code",
            "client_id": self._client_id,
            "redirect_uri": redirect_uri,
            "state": state,
        }
        if code_challenge:
            params["code_challenge"] = code_challenge
        if code_challenge_method:
            params["code_challenge_method"] = code_challenge_method

        uri = f"{self._get_www_base_url}/api/permission/oauth2/authorize"
        if workspace_id:
            uri = f"{self._get_www_base_url}/api/permission/oauth2/workspace_id/{workspace_id}/authorize"
        return uri + "?" + "&".join([f"{k}={quote_plus(v)}" for k, v in params.items()])

    def _refresh_access_token(self, refresh_token: str, secret: str = "") -> OAuthToken:
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {secret}"} if secret else {}
        body = {
            "grant_type": "refresh_token",
            "client_id": self._client_id,
            "refresh_token": refresh_token,
        }
        return self._requester.request("post", url, False, OAuthToken, headers=headers, body=body)

    async def _arefresh_access_token(self, refresh_token: str, secret: str = "") -> OAuthToken:
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {secret}"} if secret else {}
        body = {
            "grant_type": "refresh_token",
            "client_id": self._client_id,
            "refresh_token": refresh_token,
        }
        return await self._requester.arequest("post", url, False, OAuthToken, headers=headers, body=body)

    @property
    def _get_www_base_url(self) -> str:
        if self._www_base_url:
            return self._www_base_url
        if self._base_url in [COZE_CN_BASE_URL, COZE_COM_BASE_URL]:
            return self._base_url.replace("api", "www")
        return self._base_url


class WebOAuthApp(OAuthApp):
    """
    Normal OAuth App.
    """

    def __init__(self, client_id: str, client_secret: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        """
        :param client_id:
        :param client_secret:
        :param base_url:
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._base_url = remove_url_trailing_slash(base_url)
        self._api_endpoint = urlparse(base_url).netloc
        self._token = ""
        super().__init__(client_id, base_url, www_base_url=www_base_url)

    def get_oauth_url(
        self,
        redirect_uri: str,
        state: str = "",
        workspace_id: Optional[str] = None,
    ):
        """
        Get the pkce flow authorized url.

        :param redirect_uri: The redirect_uri of your app, where authentication responses can be sent and received by
        your app. It must exactly match one of the redirect URIs you registered in the OAuth Apps.
        :param state: A value included in the request that is also returned in the token response. It can be a string
        of any hash value.
        :param workspace_id:
        :return:
        """
        return self._get_oauth_url(
            redirect_uri,
            state=state,
            workspace_id=workspace_id,
        )

    def get_access_token(
        self,
        redirect_uri: str,
        code: str,
    ) -> OAuthToken:
        """
        Get the token by jwt with jwt auth flow.
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {self._client_secret}"}
        body = {
            "client_id": self._client_id,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_uri,
        }
        return self._requester.request("post", url, False, OAuthToken, headers=headers, body=body)

    def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return self._refresh_access_token(refresh_token, self._client_secret)


class AsyncWebOAuthApp(OAuthApp):
    """
    Normal OAuth App.
    """

    def __init__(self, client_id: str, client_secret: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        """
        :param client_id:
        :param client_secret:
        :param base_url:
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._base_url = remove_url_trailing_slash(base_url)
        self._api_endpoint = urlparse(base_url).netloc
        self._token = ""
        super().__init__(client_id, base_url, www_base_url=www_base_url)

    def get_oauth_url(
        self,
        redirect_uri: str,
        state: str = "",
        workspace_id: Optional[str] = None,
    ):
        """
        Get the pkce flow authorized url.

        :param redirect_uri: The redirect_uri of your app, where authentication responses can be sent and received by
        your app. It must exactly match one of the redirect URIs you registered in the OAuth Apps.
        :param state: A value included in the request that is also returned in the token response. It can be a string
        of any hash value.
        :param workspace_id:
        :return:
        """
        return self._get_oauth_url(
            redirect_uri,
            state=state,
            workspace_id=workspace_id,
        )

    async def get_access_token(
        self,
        redirect_uri: str,
        code: str,
    ) -> OAuthToken:
        """
        Get the token by jwt with jwt auth flow.
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {self._client_secret}"}
        body = {
            "client_id": self._client_id,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_uri,
        }
        return await self._requester.arequest("post", url, False, OAuthToken, headers=headers, body=body)

    async def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return await self._arefresh_access_token(refresh_token, self._client_secret)


class JWTOAuthApp(OAuthApp):
    """
    JWT OAuth App.
    """

    def __init__(self, client_id: str, private_key: str, public_key_id: str, base_url: str = COZE_COM_BASE_URL):
        """
        :param client_id:
        :param private_key:
        :param public_key_id:
        :param base_url:
        """
        self._client_id = client_id
        self._base_url = remove_url_trailing_slash(base_url)
        self._api_endpoint = urlparse(base_url).netloc
        self._token = ""
        self._private_key = private_key
        self._public_key_id = public_key_id
        super().__init__(client_id, base_url, www_base_url="")

    def get_access_token(
        self, ttl: int = 900, scope: Optional[Scope] = None, session_name: Optional[str] = None
    ) -> OAuthToken:
        """
        Get the token by jwt with jwt auth flow.

        :param ttl: The validity period of the AccessToken you apply for is in seconds.
        The default value is 900 seconds and the maximum value you can set is 86399 seconds,
        which is 24 hours.
        :param scope:
        :param session_name: Isolate different sub-resources under the same jwt account
        """
        jwt_token = self._gen_jwt(3600, session_name)
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {jwt_token}"}
        body = {
            "duration_seconds": ttl,
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "scope": scope.model_dump() if scope else None,
        }
        return self._requester.request("post", url, False, OAuthToken, headers=headers, body=body)

    def _gen_jwt(self, ttl: int, session_name: Optional[str] = None):
        now = int(time.time())
        header = {"alg": "RS256", "typ": "JWT", "kid": self._public_key_id}
        payload = {
            "iss": self._client_id,
            "aud": self._api_endpoint,
            "iat": now,
            "exp": now + ttl,
            "jti": random_hex(16),
        }
        if session_name:
            payload["session_name"] = session_name
        s = jwt.encode(header, payload, self._private_key)
        return s.decode("utf-8")


class AsyncJWTOAuthApp(OAuthApp):
    """
    JWT OAuth App.
    """

    def __init__(self, client_id: str, private_key: str, public_key_id: str, base_url: str = COZE_COM_BASE_URL):
        """
        :param client_id:
        :param private_key:
        :param public_key_id:
        :param base_url:
        """
        self._client_id = client_id
        self._base_url = remove_url_trailing_slash(base_url)
        self._api_endpoint = urlparse(base_url).netloc
        self._token = ""
        self._private_key = private_key
        self._public_key_id = public_key_id
        super().__init__(client_id, base_url, www_base_url="")

    async def get_access_token(self, ttl: int, scope: Optional[Scope] = None) -> OAuthToken:
        """
        Get the token by jwt with jwt auth flow.
        """
        jwt_token = self._gen_jwt(3600)
        url = f"{self._base_url}/api/permission/oauth2/token"
        headers = {"Authorization": f"Bearer {jwt_token}"}
        body = {
            "duration_seconds": ttl,
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "scope": scope.model_dump() if scope else None,
        }
        return await self._requester.arequest("post", url, False, OAuthToken, headers=headers, body=body)

    def _gen_jwt(self, ttl: int):
        now = int(time.time())
        header = {"alg": "RS256", "typ": "JWT", "kid": self._public_key_id}
        payload = {
            "iss": self._client_id,
            "aud": self._api_endpoint,
            "iat": now,
            "exp": now + ttl,
            "jti": random_hex(16),
        }
        s = jwt.encode(header, payload, self._private_key)
        return s.decode("utf-8")


class PKCEOAuthApp(OAuthApp):
    """
    PKCE OAuth App.
    """

    def __init__(self, client_id: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        self._token = ""
        self._requester = Requester()
        super().__init__(
            client_id,
            base_url,
            www_base_url,
        )

    def get_oauth_url(
        self,
        redirect_uri: str,
        code_verifier: str,
        code_challenge_method: Literal["plain", "S256"] = "plain",
        state: str = "",
        workspace_id: Optional[str] = None,
    ):
        """
        Get the pkce flow authorized url.

        :param redirect_uri: The redirect_uri of your app, where authentication responses can be sent and received by
        your app. It must exactly match one of the redirect URIs you registered in the OAuth Apps.
        :param code_verifier:
        :param code_challenge_method:
        :param state: A value included in the request that is also returned in the token response. It can be a string
        of any hash value.
        :param workspace_id:
        :return:
        """
        code_challenge = code_verifier if code_challenge_method == "plain" else gen_s256_code_challenge(code_verifier)

        return self._get_oauth_url(
            redirect_uri,
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method,
            state=state,
            workspace_id=workspace_id,
        )

    def get_access_token(self, redirect_uri: str, code: str, code_verifier: str) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param redirect_uri:
        :param code_verifier:
        :param code:
        :return:
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        body = {
            "grant_type": "authorization_code",
            "code": code,
            "client_id": self._client_id,
            "redirect_uri": redirect_uri,
            "code_verifier": code_verifier,
        }
        return self._requester.request("post", url, False, OAuthToken, body=body)

    def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return self._refresh_access_token(refresh_token)


class AsyncPKCEOAuthApp(OAuthApp):
    """
    PKCE OAuth App.
    """

    def __init__(self, client_id: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        self._token = ""
        self._requester = Requester()
        super().__init__(
            client_id,
            base_url,
            www_base_url,
        )

    def get_oauth_url(
        self,
        redirect_uri: str,
        code_verifier: str,
        code_challenge_method: Literal["plain", "S256"] = "plain",
        state: str = "",
        workspace_id: Optional[str] = None,
    ):
        """
        Get the pkce flow authorized url.

        :param redirect_uri: The redirect_uri of your app, where authentication responses can be sent and received by
        your app. It must exactly match one of the redirect URIs you registered in the OAuth Apps.
        :param code_verifier:
        :param code_challenge_method:
        :param state: A value included in the request that is also returned in the token response. It can be a string
        of any hash value.
        :param workspace_id:
        :return:
        """
        code_challenge = code_verifier if code_challenge_method == "plain" else gen_s256_code_challenge(code_verifier)

        return self._get_oauth_url(
            redirect_uri,
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method,
            state=state,
            workspace_id=workspace_id,
        )

    async def get_access_token(self, redirect_uri: str, code: str, code_verifier: str) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param redirect_uri:
        :param code_verifier:
        :param code:
        :return:
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        body = {
            "grant_type": "authorization_code",
            "code": code,
            "client_id": self._client_id,
            "redirect_uri": redirect_uri,
            "code_verifier": code_verifier,
        }
        return await self._requester.arequest("post", url, False, OAuthToken, body=body)

    async def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return await self._arefresh_access_token(refresh_token)


class DeviceOAuthApp(OAuthApp):
    """
    Device OAuth App.
    """

    def __init__(self, client_id: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        self._token = ""
        self._requester = Requester()
        super().__init__(
            client_id,
            base_url,
            www_base_url,
        )

    def get_device_code(
        self,
        workspace_id: Optional[str] = None,
    ) -> DeviceAuthCode:
        """
        Get the pkce flow authorized url.

        :param workspace_id:
        :return:
        """

        uri = f"{self._base_url}/api/permission/oauth2/device/code"
        if workspace_id:
            uri = f"{self._base_url}/api/permission/oauth2/workspace_id/{workspace_id}/device/code"
        body = {
            "client_id": self._client_id,
        }
        headers = {
            "Content-Type": "application/json",
        }
        res = self._requester.request("post", uri, False, DeviceAuthCode, headers=headers, body=body)
        res.verification_url = f"{res.verification_uri}?user_code={res.user_code}"
        return res

    def get_access_token(self, device_code: str, poll: bool = False) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param device_code:
        :param poll: whether to poll for the token
        :return:
        """
        if not poll:
            return self._get_access_token(device_code)

        interval = 5
        while True:
            try:
                return self._get_access_token(device_code)
            except CozePKCEAuthError as e:
                if e.error == CozePKCEAuthErrorType.AUTHORIZATION_PENDING:
                    time.sleep(interval)
                    continue
                elif e.error == CozePKCEAuthErrorType.SLOW_DOWN:
                    if interval < 30:
                        interval += 5
                    time.sleep(interval)
                    continue
                else:
                    raise

    def _get_access_token(self, device_code: str, poll: bool = False) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param device_code:
        :param poll: whether to poll for the token
        :return:
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        body = {
            "client_id": self._client_id,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
        }
        return self._requester.request("post", url, False, OAuthToken, body=body)

    def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return self._refresh_access_token(refresh_token)


class AsyncDeviceOAuthApp(OAuthApp):
    """
    Device OAuth App.
    """

    def __init__(self, client_id: str, base_url: str = COZE_COM_BASE_URL, www_base_url: str = ""):
        self._token = ""
        self._requester = Requester()
        super().__init__(
            client_id,
            base_url,
            www_base_url,
        )

    async def get_device_code(self, workspace_id: Optional[str] = None) -> DeviceAuthCode:
        """
        Get the pkce flow authorized url.

        :param workspace_id:
        :return:
        """

        uri = f"{self._base_url}/api/permission/oauth2/device/code"
        if workspace_id:
            uri = f"{self._base_url}/api/permission/oauth2/workspace_id/{workspace_id}/device/code"
        body = {
            "client_id": self._client_id,
        }
        headers = {
            "Content-Type": "application/json",
        }
        res = await self._requester.arequest("post", uri, False, DeviceAuthCode, headers=headers, body=body)
        res.verification_url = f"{res.verification_uri}?user_code={res.user_code}"
        return res

    async def get_access_token(self, device_code: str, poll: bool = False) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param device_code:
        :param poll: whether to poll for the token
        :return:
        """
        if not poll:
            return await self._get_access_token(device_code)

        interval = 5
        while True:
            try:
                return await self._get_access_token(device_code)
            except CozePKCEAuthError as e:
                if e.error == CozePKCEAuthErrorType.AUTHORIZATION_PENDING:
                    time.sleep(interval)
                    continue
                elif e.error == CozePKCEAuthErrorType.SLOW_DOWN:
                    if interval < 30:
                        interval += 5
                    time.sleep(interval)
                    continue
                else:
                    raise

    async def _get_access_token(self, device_code: str, poll: bool = False) -> OAuthToken:
        """
        Get the token with pkce auth flow.

        :param device_code:
        :param poll: whether to poll for the token
        :return:
        """
        url = f"{self._base_url}/api/permission/oauth2/token"
        body = {
            "client_id": self._client_id,
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
        }
        return await self._requester.arequest("post", url, False, OAuthToken, body=body)

    async def refresh_access_token(self, refresh_token: str) -> OAuthToken:
        return await self._arefresh_access_token(refresh_token)


class Auth(abc.ABC):
    """
    This class is the base class for all authorization types.

    It provides the abstract methods for getting the token type and token.
    """

    @property
    @abc.abstractmethod
    def token_type(self) -> str:
        """
        The authorization type used in the http request header.

        eg: Bearer, Basic, etc.

        :return: token type
        """

    @property
    @abc.abstractmethod
    def token(self) -> str:
        """
        The token used in the http request header.

        :return: token
        """

    def authentication(self, headers: dict) -> None:
        """
        Construct the authorization header in the http headers.

        :param headers: http headers
        :return: None
        """
        headers["Authorization"] = f"{self.token_type} {self.token}"


class TokenAuth(Auth):
    """
    The fixed access token auth flow.
    """

    def __init__(self, token: str):
        assert isinstance(token, str)
        assert len(token) > 0
        self._token = token

    @property
    def token_type(self) -> str:
        return "Bearer"

    @property
    def token(self) -> str:
        return self._token


class JWTAuth(Auth):
    """
    The JWT auth flow.
    """

    def __init__(
        self,
        client_id: str,
        private_key: str,
        public_key_id: str,
        ttl: int = 7200,
        base_url: str = COZE_COM_BASE_URL,
    ):
        assert isinstance(client_id, str)
        assert isinstance(private_key, str)
        assert isinstance(public_key_id, str)
        assert isinstance(ttl, int)
        assert ttl > 0
        assert isinstance(base_url, str)

        self._client_id = client_id
        self._ttl = ttl
        self._base_url = remove_url_trailing_slash(base_url)
        self._token = None
        self._oauth_cli = JWTOAuthApp(self._client_id, private_key, public_key_id, base_url=self._base_url)

    @property
    def token_type(self) -> str:
        return "Bearer"

    @property
    def token(self) -> str:
        token = self._generate_token()
        return token.access_token

    def _generate_token(self):
        if self._token is not None and int(time.time()) < self._token.expires_in:
            return self._token
        self._token = self._oauth_cli.get_access_token(self._ttl)
        return self._token



from typing import Optional

from cozepy.auth import Auth
from cozepy.model import CozeModel
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash


class CreateRoomResult(CozeModel):
    # Token to join the room
    token: str
    # The id of user
    uid: str
    # The id of room
    room_id: str
    # App id to join the room
    app_id: str


class RoomsClient(object):
    """
    Room service client.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    def create(
        self, *, bot_id: str, voice_id: Optional[str] = None, conversation_id: Optional[str] = None
    ) -> CreateRoomResult:
        """
        create rtc room

        :param bot_id: The id of the bot.
        :param voice_id: The voice id of the voice.
        :param conversation_id: The id of the conversation.
        :return: create room result
        """
        url = f"{self._base_url}/v1/audio/rooms"
        body = {
            "bot_id": bot_id,
            "voice_id": voice_id,
            "conversation_id": conversation_id,
        }
        return self._requester.request("post", url, stream=False, cast=CreateRoomResult, body=body)


class AsyncRoomsClient(object):
    """
    Room service async client.
    """

    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

    async def create(
        self, *, bot_id: str, voice_id: Optional[str] = None, conversation_id: Optional[str] = None
    ) -> CreateRoomResult:
        """
        create rtc room

        :param bot_id: The id of the bot.
        :param voice_id: The voice id of the voice.
        :return:
        """
        url = f"{self._base_url}/v1/audio/rooms"
        body = {
            "bot_id": bot_id,
            "voice_id": voice_id,
            "conversation_id": conversation_id,
        }
        return await self._requester.arequest("post", url, stream=False, cast=CreateRoomResult, body=body)



from typing import TYPE_CHECKING, Optional

from cozepy.auth import Auth
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash

if TYPE_CHECKING:
    from .rooms import AsyncRoomsClient, RoomsClient
    from .speech import AsyncSpeechClient, SpeechClient
    from .voices import AsyncVoicesClient, VoicesClient


class AudioClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

        self._rooms: Optional[RoomsClient] = None
        self._voices: Optional[VoicesClient] = None
        self._speech: Optional[SpeechClient] = None

    @property
    def rooms(self) -> "RoomsClient":
        if self._rooms is None:
            from .rooms import RoomsClient

            self._rooms = RoomsClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._rooms

    @property
    def speech(self) -> "SpeechClient":
        if self._speech is None:
            from .speech import SpeechClient

            self._speech = SpeechClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._speech

    @property
    def voices(self) -> "VoicesClient":
        if self._voices is None:
            from .voices import VoicesClient

            self._voices = VoicesClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._voices


class AsyncAudioClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester

        self._rooms: Optional[AsyncRoomsClient] = None
        self._voices: Optional[AsyncVoicesClient] = None
        self._speech: Optional[AsyncSpeechClient] = None

    @property
    def rooms(self) -> "AsyncRoomsClient":
        if self._rooms is None:
            from .rooms import AsyncRoomsClient

            self._rooms = AsyncRoomsClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._rooms

    @property
    def speech(self) -> "AsyncSpeechClient":
        if self._speech is None:
            from .speech import AsyncSpeechClient

            self._speech = AsyncSpeechClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._speech

    @property
    def voices(self) -> "AsyncVoicesClient":
        if self._voices is None:
            from .voices import AsyncVoicesClient

            self._voices = AsyncVoicesClient(base_url=self._base_url, auth=self._auth, requester=self._requester)
        return self._voices



from typing import Any, Dict, List, Optional

from cozepy.auth import Auth
from cozepy.chat import Message
from cozepy.model import AsyncNumberPaged, CozeModel, HTTPRequest, NumberPaged
from cozepy.request import Requester
from cozepy.util import remove_url_trailing_slash


class Conversation(CozeModel):
    id: str
    created_at: int
    meta_data: Dict[str, str]
    # section_id is used to distinguish the context sections of the session history. The same section is one context.
    last_section_id: str


class Section(CozeModel):
    id: str
    conversation_id: str


class _PrivateListConversationResp(CozeModel):
    has_more: bool
    conversations: List[Conversation]

    def get_total(self) -> Optional[int]:
        return None

    def get_has_more(self) -> Optional[bool]:
        return self.has_more

    def get_items(self) -> List[Conversation]:
        return self.conversations


class ConversationsClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._messages = None

    def create(
        self,
        *,
        messages: Optional[List[Message]] = None,
        bot_id: Optional[str] = None,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Conversation:
        """
        Create a conversation.
        Conversation is an interaction between a bot and a user, including one or more messages.

        docs en: https://www.coze.com/docs/developer_guides/create_conversation
        docs zh: https://www.coze.cn/docs/developer_guides/create_conversation

        :param messages: Messages in the conversation. For more information, see EnterMessage object.
        :param bot_id: Bind and isolate conversation on different bots.
        :param meta_data: Additional information when creating a message, and this additional information will also be
        returned when retrieving messages.
        :return: Conversation object
        """
        url = f"{self._base_url}/v1/conversation/create"
        body: Dict[str, Any] = {
            "messages": [i.model_dump() for i in messages] if messages and len(messages) > 0 else [],
            "meta_data": meta_data,
        }
        if bot_id:
            body["bot_id"] = bot_id
        return self._requester.request("post", url, False, Conversation, body=body)

    def list(
        self,
        *,
        bot_id: str,
        page_num: int = 1,
        page_size: int = 50,
    ):
        url = f"{self._base_url}/v1/conversations"

        def request_maker(i_page_num: int, i_page_size: int) -> HTTPRequest:
            return self._requester.make_request(
                "GET",
                url,
                params={
                    "bot_id": bot_id,
                    "page_num": i_page_num,
                    "page_size": i_page_size,
                },
                cast=_PrivateListConversationResp,
                is_async=False,
                stream=False,
            )

        return NumberPaged(
            page_num=page_num,
            page_size=page_size,
            requestor=self._requester,
            request_maker=request_maker,
        )

    def retrieve(self, *, conversation_id: str) -> Conversation:
        """
        Get the information of specific conversation.

        docs en: https://www.coze.com/docs/developer_guides/retrieve_conversation
        docs cn: https://www.coze.cn/docs/developer_guides/retrieve_conversation

        :param conversation_id: The ID of the conversation.
        :return: Conversation object
        """
        url = f"{self._base_url}/v1/conversation/retrieve"
        params = {
            "conversation_id": conversation_id,
        }
        return self._requester.request("get", url, False, Conversation, params=params)

    def clear(self, *, conversation_id: str) -> Section:
        url = f"{self._base_url}/v1/conversations/{conversation_id}/clear"
        return self._requester.request("post", url, False, Section)

    @property
    def messages(self):
        if not self._messages:
            from .message import MessagesClient

            self._messages = MessagesClient(self._base_url, self._auth, self._requester)
        return self._messages


class AsyncConversationsClient(object):
    def __init__(self, base_url: str, auth: Auth, requester: Requester):
        self._base_url = remove_url_trailing_slash(base_url)
        self._auth = auth
        self._requester = requester
        self._messages = None

    async def create(
        self,
        *,
        messages: Optional[List[Message]] = None,
        bot_id: Optional[str] = None,
        meta_data: Optional[Dict[str, str]] = None,
    ) -> Conversation:
        """
        Create a conversation.
        Conversation is an interaction between a bot and a user, including one or more messages.

        docs en: https://www.coze.com/docs/developer_guides/create_conversation
        docs zh: https://www.coze.cn/docs/developer_guides/create_conversation

        :param messages: Messages in the conversation. For more information, see EnterMessage object.
        :param bot_id: Bind and isolate conversation on different bots.
        :param meta_data: Additional information when creating a message, and this additional information will also be
        returned when retrieving messages.
        :return: Conversation object
        """
        url = f"{self._base_url}/v1/conversation/create"
        body: Dict[str, Any] = {
            "messages": [i.model_dump() for i in messages] if messages and len(messages) > 0 else [],
            "meta_data": meta_data,
        }
        if bot_id:
            body["bot_id"] = bot_id
        return await self._requester.arequest("post", url, False, Conversation, body=body)

    async def list(
        self,
        *,
        bot_id: str,
        page_num: int = 1,
        page_size: int = 50,
    ):
        url = f"{self._base_url}/v1/conversations"

        def request_maker(i_page_num: int, i_page_size: int) -> HTTPRequest:
            return self._requester.make_request(
                "GET",
                url,
                params={
                    "bot_id": bot_id,
                    "page_num": i_page_num,
                    "page_size": i_page_size,
                },
                cast=_PrivateListConversationResp,
                is_async=False,
                stream=False,
            )

        return await AsyncNumberPaged.build(
            page_num=page_num,
            page_size=page_size,
            requestor=self._requester,
            request_maker=request_maker,
        )

    async def retrieve(self, *, conversation_id: str) -> Conversation:
        """
        Get the information of specific conversation.

        docs en: https://www.coze.com/docs/developer_guides/retrieve_conversation
        docs cn: https://www.coze.cn/docs/developer_guides/retrieve_conversation

        :param conversation_id: The ID of the conversation.
        :return: Conversation object
        """
        url = f"{self._base_url}/v1/conversation/retrieve"
        params = {
            "conversation_id": conversation_id,
        }
        return await self._requester.arequest("get", url, False, Conversation, params=params)

    async def clear(self, *, conversation_id: str) -> Section:
        url = f"{self._base_url}/v1/conversations/{conversation_id}/clear"
        return await self._requester.arequest("post", url, False, Section)

    @property
    def messages(self):
        if not self._messages:
            from .message import AsyncMessagesClient

            self._messages = AsyncMessagesClient(self._base_url, self._auth, self._requester)
        return self._messages
